#!/usr/bin/env python3
#
#    ts590-monitor.py
#
#    Simple SWR meter monitor.
#
#    Copyright (C) 2023 by Matt Roberts.
#    License: GNU GPL3 (www.gnu.org)
#
#

# system modules
import sys
import os
import os.path
import time
import json
import curses
import threading
import datasource
import _thread
from datetime import datetime

# color pairs
WHITE_ON_CLEAR = 1
YELLOW_ON_CLEAR = 2
GREEN_ON_CLEAR = 3
RED_ON_CLEAR = 4
BLUE_ON_CLEAR = 5

BLACK_ON_WHITE = 6
BLACK_ON_YELLOW = 7
BLACK_ON_GREEN = 8
BLACK_ON_RED = 9
BLACK_ON_BLUE = 10

WHITE_ON_RED = 11
WHITE_ON_BLUE = 12
WHITE_ON_GREEN = 13
WHITE_ON_YELLOW = 14

YELLOW_ON_BLUE = 15


# enable exception display
ShowExceptions = False

# meter state
MeterState = { }


#
#  initialize()
#
def initialize(stdscr):
	stdscr.clear()
	curses.start_color()
	curses.use_default_colors()
	curses.cbreak()
	curses.noecho()
	curses.curs_set(0)
	stdscr.keypad(True)
	stdscr.notimeout(True)
	curses.halfdelay(2) # 0.2s

	# normal colors
	curses.init_pair(WHITE_ON_CLEAR, curses.COLOR_WHITE, -1) # w/bk
	curses.init_pair(YELLOW_ON_CLEAR, curses.COLOR_YELLOW, -1) # y/bk
	curses.init_pair(GREEN_ON_CLEAR, curses.COLOR_GREEN, -1) # g/bk
	curses.init_pair(RED_ON_CLEAR, curses.COLOR_RED, -1) # r/bk
	curses.init_pair(BLUE_ON_CLEAR, curses.COLOR_BLUE, -1) # r/bk
	# inverted colors
	curses.init_pair(BLACK_ON_WHITE, curses.COLOR_BLACK, curses.COLOR_WHITE) # bk/w
	curses.init_pair(BLACK_ON_YELLOW, curses.COLOR_BLACK, curses.COLOR_YELLOW) # bk/y
	curses.init_pair(BLACK_ON_GREEN, curses.COLOR_BLACK, curses.COLOR_GREEN) # bk/g
	curses.init_pair(BLACK_ON_RED, curses.COLOR_BLACK, curses.COLOR_RED) # bk/r
	curses.init_pair(BLACK_ON_BLUE, curses.COLOR_BLACK, curses.COLOR_BLUE) # bk/bl
	# white on colors
	curses.init_pair(WHITE_ON_RED, curses.COLOR_WHITE, curses.COLOR_RED) # wh/rd
	curses.init_pair(WHITE_ON_BLUE, curses.COLOR_WHITE, curses.COLOR_BLUE) # wh/blu
	curses.init_pair(WHITE_ON_GREEN, curses.COLOR_WHITE, curses.COLOR_GREEN) # wh/blu
	curses.init_pair(WHITE_ON_YELLOW, curses.COLOR_WHITE, curses.COLOR_YELLOW) # wh/y
	# yellow on colors
	curses.init_pair(YELLOW_ON_BLUE, curses.COLOR_YELLOW, curses.COLOR_BLUE) # wh/blu


#
#  drawmeter(dots) - draw a 30-dot meter
#
def drawmeter(stdscr, dots, bar_style = 0, stencil_style = 0):
	# default styles
	if not bar_style:
		bar_style = curses.color_pair(BLACK_ON_WHITE)
	if not stencil_style:
		stencil_style = curses.color_pair(WHITE_ON_CLEAR)

	# draw...
	stdscr.addstr('[', stencil_style);
	for i in range(30):
		if i < dots:
			stdscr.addstr(' ', bar_style)
		else:
			stdscr.addstr(' ', curses.color_pair(WHITE_ON_CLEAR))
	stdscr.addstr(']', stencil_style);


#
#  getsunits(dots) return approximate S-units corresponding to meter dots
#
def getsunits(dots):
	if dots > 15:
		return "S9+%02d" % int((dots - 15) * (60.0 / 15.0))
	return " S-%d" % int(dots * (9.0 / 15.0))


#
#  getswr(dots) - return approximate SWR corresponding to meter dots
#
def getswr(dots):
	if dots > 15:
		return "%0.1f" % (3.0 + (0.5 * (dots - 16)))
	if dots > 10:
		return "%0.1f" % (2.0 + (0.2 * (dots - 11)))
	if dots > 1:
		return "%0.1f" % (1.0 + (0.1 * (dots - 1)))
	if dots == 1:
		return "1.0"
	return ""


#
#  formatfreq(...)
#
def formatfreq(f):
	return ('{:,d}'.format(f)).replace(',', '.')


#
#  main()
#
def main(stdscr):
	global ShowExceptions
	global MeterState

	# setup
	initialize(stdscr)

	# tickle the update logic (TODO: this shouldn't be necessary)
	ch = stdscr.getch()

	# initialize the spinner icon
	spinner = False

	# build the style for normal state icons
	icon_style = curses.color_pair(WHITE_ON_CLEAR) | curses.A_BOLD

	# wait for a keypress
	while True:
		#
		#  Sample Data:
		#    
		#   {
		#		"fa": 28074000,
		#		"tx": false,
		#		"split": false,
		#		"fb": 28074500,
		#		"mode": "USB",
		#		"meter": 16,
		#		"swr": 0,
		#		"comp": 0,
		#		"alc": 0
		#	}
		#
		#
		if ('init' not in MeterState.keys()) or MeterState['init']:
			row, col = stdscr.getmaxyx()
			row = int(row / 2)
			col = int((col / 2) - 7)
			if row < 0:
				row = 0
			if col < 0:
				col = 0
			stdscr.addstr(row, col, "Initializing...")
			stdscr.refresh()
		else:
			row, col = stdscr.getmaxyx()
			row = int((row / 2) - 2)
			col = int((col / 2) - 22)
			if row < 0:
				row = 0
			if col < 0:
				col = 0

			# Draw VFO A
			stdscr.addstr(row, col, "VfoA: ")
			stdscr.addstr(row, col + 8, '%11s' % formatfreq(MeterState['fa']), curses.A_BOLD)
			if MeterState['split'] and not MeterState['tx']:
				stdscr.addstr(" <")
			stdscr.clrtoeol()

			# Draw Activity Spinner
			spinner = not spinner
			stdscr.addstr(row, col + 6, "\u00b7" if spinner else " ",
				curses.color_pair(WHITE_ON_CLEAR) | curses.A_BOLD)

			# Draw Preamp/Attenuator (one/both/off)
			pre = MeterState['pre']
			att = MeterState['att']
			if pre or att:
				if pre and att:
					icon = '[P/A]'
				elif pre:
					icon = '[PRE]'
				else:
					icon = '[ATT]'
				stdscr.addstr(row, col + 25, ' ')
				stdscr.addstr(icon, icon_style)

			# Draw Mode
			icon = '[%s]' % MeterState['mode']
			pad = 6 - len(icon)
			stdscr.addstr(row, col + 31, ' ')
			for i in range(pad):
				stdscr.addstr(' ')
			style = curses.color_pair(WHITE_ON_YELLOW) | curses.A_BOLD
			if MeterState['mode'].endswith('-D'):
				style = curses.color_pair(WHITE_ON_BLUE) | curses.A_BOLD
			stdscr.addstr(icon, style)

			# Draw TX/RX
			if MeterState['tx']:
				stdscr.addstr(row, col + 40, "[TX]", curses.color_pair(WHITE_ON_RED) | curses.A_BOLD)
			else:
				stdscr.addstr(row, col + 40, "[RX]", curses.color_pair(WHITE_ON_GREEN) | curses.A_BOLD)
			row += 1

			# Draw VFO B
			style = 0 if MeterState['split'] else curses.A_DIM
			stdscr.addstr(row, col, "VfoB: ", style)
			style = curses.A_BOLD if MeterState['split'] else curses.A_DIM
			stdscr.addstr(row, col + 8, '%11s' % formatfreq(MeterState['fb']), style)
			if MeterState['split'] and MeterState['tx']:
				stdscr.addstr(" <")
			stdscr.clrtoeol()

			# Draw Noise Blankers (off/1/2)
			if MeterState['nb']:
				icon = '[NB1]' if MeterState['nb'] == 1 else '[NB2]'
				stdscr.addstr(row, col + 22, ' ')
				stdscr.addstr(icon, icon_style)

			# Draw Antenna Tuner (on/off)
			if MeterState['atu']:
				icon = '[ATU]'
				stdscr.addstr(row, col + 28, ' ')
				stdscr.addstr(icon, icon_style)

			# Draw RX Antenna (on/off)
			if MeterState['rxant']:
				icon = '[RX]'
				stdscr.addstr(row, col + 34, ' ')
				stdscr.addstr(icon, icon_style)

			# Draw Main Antenna (1/2)
			icon = '[A%d]' % MeterState['ant']
			stdscr.addstr(row, col + 39, ' ')
			stdscr.addstr(icon, icon_style)
			stdscr.clrtoeol()
			row += 1

			# Draw S/PO Meter
			stdscr.addstr(row, col, "S/PO: ")
			drawmeter(stdscr, MeterState['meter'])
			stdscr.addstr(" ")
			if MeterState['tx']:
				pct = int(round(100.0 * MeterState['meter'] / 30.0))
				stdscr.addstr("%3d W" % pct, curses.A_BOLD)
			elif MeterState['meter']:
				# show S-units
				sm = getsunits(MeterState['meter'])
				# show meter percentage
				#sm = "%3d %%" % int(100.0 * MeterState['meter'] / 30.0)
				stdscr.addstr(sm, curses.A_BOLD)
			else:
				stdscr.addstr(" \u00b7\u00b7\u00b7", curses.A_DIM);
			stdscr.clrtoeol()
			row += 1

			# dim labels for SWR and ALC when not transmitting
			stencil_style = 0 if MeterState['tx'] else curses.A_DIM

			# Draw SWR Meter
			stdscr.addstr(row, col, "SWR : ", stencil_style)
			swr = getswr(MeterState['swr'])
			style = 0
			if swr:
				dswr = float(swr) # swr to one decimal point
				if dswr >= 2:
					style = curses.color_pair(BLACK_ON_RED)
				elif dswr >= 1.5:
					style = curses.color_pair(BLACK_ON_YELLOW)
				else:
					style = curses.color_pair(BLACK_ON_GREEN)
			drawmeter(stdscr, MeterState['swr'], style, stencil_style)
			if swr:
				stdscr.addstr(" ")
				stdscr.addstr(swr, curses.A_BOLD)
				stdscr.addstr(":1", curses.A_BOLD)
			else:
				stdscr.addstr("  \u00b7\u00b7\u00b7", curses.A_DIM);
			stdscr.clrtoeol()
			row += 1

			# Draw ALC Meter
			stdscr.addstr(row, col, "ALC : ", stencil_style)
			style = 0
			pct = int(200 * MeterState['alc'] / 30.0)
			if pct > 100:
				style = curses.color_pair(BLACK_ON_RED)
			elif pct > 50:
				style = curses.color_pair(BLACK_ON_YELLOW)
			else:
				style = curses.color_pair(BLACK_ON_GREEN)
			drawmeter(stdscr, MeterState['alc'], style, stencil_style)
			if MeterState['tx']:
				stdscr.addstr(" %3d %%" % pct, curses.A_BOLD)
			else:
				stdscr.addstr("  \u00b7\u00b7\u00b7", curses.A_DIM);
			stdscr.clrtoeol()
			row += 1
			stdscr.refresh()

		# read key input
		ch = None
		try:
			ch = stdscr.getch()
		except TimeoutError as ex:
			# no key available
			time.sleep(0.1)
			continue

		# handle keystroke
		if ch == curses.KEY_RESIZE: # SIGWINCH
			stdscr.clear()
		elif ch == ord('q'):
			break
		elif ch == ord('r'):
			stdscr.refresh()

	# shut down curses
	curses.endwin()


#
#  data update callback
#
def callback(data):
	global MeterState

	# DEBUG:
	#sys.stderr.write(data)
	#sys.stderr.write('\n')
	#sys.stderr.flush()

	try:
		j = json.loads(data)
		for k in j.keys():
			MeterState[k] = j[k]
	except Exception as ex:
		sys.stderr.write("Error: %s\n" % str(ex))


#
#  entry point
#
if __name__ == '__main__':
	# watch the state in a background thread
	args = sys.argv[1:]
	if not args:
		if os.path.exists("./ts590sg-state.py"):
			args = "./ts590sg-state.py"
		else:
			args = "ts590sg-state.py"

	# start the state management in its own thread
	_thread.start_new(datasource.watchstate, (args, callback))

	# run the UI
	curses.wrapper(main)

	# clean up
	datasource.stop()

# EOF
