#!/usr/bin/env python3
#
#  ts590sg-state.py
#
#  Poll critical tranceiver data from rigctld.
#
#  Copyright (c) 2023 by Matt Roberts,
#  All rights reserved.
#
#  License: GPL3
#

import socket
import select
import json
import time
import sys
import os

# minimum amount of time between message fetches
MessageInterval = 0.20 # s


#
#  return 'meter dots' in the range [0, 30]
#
def meter(message):
	if not message:
		return 0
	return int(message[5:7])


#
#  process one message
#
def process(message, state):
	# safety
	if not message:
		return False
	if state is None:
		return False
	
	# main status message
	if message.startswith('IF'):
		# read FA
		f = message[2:13]
		while f and f[0] == '0':
			f = f[1:]
		state['fa'] = int(f)

		# read TX flag
		state['tx'] = (message[28] == '1')

		# read split flag
		state['split'] = (message[32] == '1')

		# read split flag
		mode = message[29]
		D = "-D" if 'data' in state.keys() and state['data'] else ""
		if mode == '1':
			state['mode'] = 'LSB' + D
		elif mode == '2':
			state['mode'] = 'USB' + D
		elif mode == '3':
			state['mode'] = 'CW'
		elif mode == '4':
			state['mode'] = 'FM' + D
		elif mode == '5':
			state['mode'] = 'AM' + D
		elif mode == '6':
			state['mode'] = 'FSK'
		elif mode == '7':
			state['mode'] = 'CW-R'
		elif mode == '9':
			state['mode'] = 'FSK-R'
		else:
			state['mode'] = mode
	
	# VFO B
	elif message.startswith('FB'):
		f = message[2:]
		while f and f[0] == '0':
			f = f[1:]
		state['fb'] = int(f)

	# S-meter (RF-meter)
	elif message.startswith('SM0'):
		state['meter'] = meter(message)

	# SWR
	elif message.startswith('RM1'):
		state['swr'] = meter(message)

	# Comp
	elif message.startswith('RM2'):
		state['comp'] = meter(message)

	# ALC
	elif message.startswith('RM3'):
		state['alc'] = meter(message)

	# Data Mode
	elif message.startswith('DA'):
		state['data'] = (message[2] == '1')

	# Noise Blankers
	elif message.startswith('NB'):
		state['nb'] = int(message[2])

	# Attenuator
	elif message.startswith('RA'):
		state['att'] = (message[3] == '1')

	# Preamplifier
	elif message.startswith('PA'):
		state['pre'] = (message[2] == '1')

	# Antenna Tuner
	elif message.startswith('AC'):
		state['atu'] = ''
		if message[3] == '1':
			state['atu'] += 'T'
		if message[2] == '1':
			state['atu'] += 'R'

	# Antenna Connectors
	elif message.startswith('AN'):
		state['ant'] = 1 if message[2] == '1' else 2
		state['rxant'] = True if message[3] == '1' else False
		state['drv'] = True if message[4] == '1' else False

	# DEBUG:
	#sys.stderr.write("DEBUG: Message: %s\n" % message);

	# success
	return True


#
#  main()
#
def main():
	global MessageInterval

	# connect to the socket (TODO: make this configurable)
	s = socket.create_connection(('localhost', 4532))
	commands = [ [
			# group #1; read ALL every scan
			b'W IF; 38', # FA + Mode + Split
			b'W FB; 14', # FB
			b'W SM0; 8', # S-meter
			b'W RM; 24'  # TX meters
		], [
			# group #2; read ONE every scan
			b'W DA; 4', # Data mode
			b'W RA; 7', # Attenuator
			b'W PA; 5', # Preamplifier
			b'W AC; 6', # Antenna Tuner
			b'W AN; 6', # Antenna Connectors
			b'W NB; 4', # Noise Blankers
		]
	]

	# command indices
	indices = [ 0, 0 ]
	group = 0

	# data state table
	state = { 'init' : True }

	now = 0
	iobuffer = ''
	last_cmd = 0
	while True:
		r, w, e = select.select([s], [], [], 0.05)

		if r:
			# process incoming data
			data = s.recv(32)
			if data:
				data = data.decode('utf-8')
				iobuffer = iobuffer + data

				while ';' in iobuffer:
					message, iobuffer = iobuffer.split(';', 1)
					message = message.replace('\n', '')
					process(message, state)

					# DEBUG:
					if sys.argv[1:] and sys.argv[1] == '-d':
						try:
							sys.stderr.write(message)
							sys.stderr.write('\n')
							sys.stderr.flush()
						except:
							return # if stderr closed, give up

				# output JSON state
				j = json.dumps(state)
				try:
					sys.stdout.write(j)
					sys.stdout.write('\n')
					sys.stdout.flush()
				except:
					return # if stdout closed, give up

		# now cycle through commands
		now = time.time()
		if now - last_cmd >= MessageInterval:
			last_cmd = now
			command = None

			# initialization
			if state['init']:
				index = indices[group]
				command = commands[group][index]

				# DEBUG
				#sys.stderr.write("Init: command = %d:%d\n" % (group, index))

				indices[group] = (indices[group] + 1) % len(commands[group])
				if indices[group] == 0:
					group = (group + 1) % len(indices)
					if group == 0:
						state['init'] = False

			# normal runtime
			else:
				# get the next command
				command = commands[group][indices[group]]

				# DEBUG
				#sys.stderr.write("Run: command = %d:%d\n" % (group, indices[group]))

				# increment the group pointer
				indices[group] = (indices[group] + 1) % len(commands[group])

				# if the group #1 rolls over, move to group #2
				if group == 0 and indices[group] == 0:
					group = 1
				elif group == 1: # else move back to group #1
					group = 0

			# send the command
			if command:
				s.send(command + b'\r')

	s.close()

if __name__ == '__main__':
	main()

# EOF
